{-# OPTIONS -Wall #-}
{-# OPTIONS -fwarn-tabs #-}
{-# OPTIONS -fwarn-incomplete-record-updates #-}
{-# OPTIONS -fwarn-monomorphism-restriction #-}
{-# OPTIONS -fwarn-unused-do-bind #-}
{-# OPTIONS -fwarn-incomplete-patterns  #-}

{-# OPTIONS -XExplicitForAll  #-}

-- from https://github.com/moizjv/haskellParallelismMapReduce

import Data.List (sortBy, groupBy) -- generice sort and group function
import Data.List.Split (splitOn)   -- tokenizer
import Data.Function (on)          -- tool

-- on :: (b -> b -> c) -> (a -> b) -> a -> a -> c
-- (.*.) `on` f = \x y -> f x .*. f y

-- sample document
egTextString :: String
egTextString = "Zola was born in Paris in 1840"

-- sample input
egInput :: [(Int, String)]
egInput = [(0, egTextString),(1, egTextString)]

-- tokenizer
wordsep :: String -> [String]
wordsep = splitOn " "

-- add 1 to every word in a list
countOne :: [String] -> [(String, Int)]
countOne = map (\x -> (x,1))

-- mapper function : discard identifier, tokenize then count 1 for each word
mapper :: (Int,String) -> [(String,Int)]
mapper = countOne . wordsep . snd

-- sample output :
-- [ [("Zola",1),("was",1),("born",1),("in",1),("Paris",1),("in",1),("1840",1)],
--   [("Zola",1),("was",1),("born",1),("in",1),("Paris",1),("in",1),("1840",1)] ]
egOutMap :: [[(String,Int)]]
egOutMap = map mapper egInput

-- sort according to tokens
combining :: [(String,Int)] -> [(String,Int)]
combining =  sortBy (compare `on` fst)

-- once sorted, group values of identical tokens
-- NB : \l -> ( fst . head $ l , map snd l ) :: [(a, b)] -> (a, [b])
-- keep the head a-value and group all b-values.
grp :: [(String,Int)] -> [(String,[Int])]
grp = map (\l -> (fst . head $ l, map snd l)) . groupBy ((==) `on` fst) 

-- sample output : 
-- [("1840",[1,1]),("Paris",[1,1]),("Zola",[1,1]),("born",[1,1]),("in",[1,1,1,1]),("was",[1,1])]
egOutGrp :: [(String,[Int])]
egOutGrp = grp . combining  . concat $ egOutMap 

-- reducer function : simply add lists of 1s using foldr function
red :: (String,[Int]) -> (String, Int)
red (s,ls) = (s, foldr1 (+) ls)

-- definition of foldr :
---- if the list is empty, the result is the initial value z; else
---- apply f to the first element and the result of folding the rest
--foldr f z []     = z 
--foldr f z (x:xs) = f x (foldr f z xs) 
 
-- complete mapreduce pipeline
mapreduce :: [(Int, String)] -> [(String, Int)]
mapreduce = map red . (grp . combining)  . concat . map mapper

-- complete mapreduce pipeline with local aggregation
mapreduce' :: [(Int, String)] -> [(String, Int)]
mapreduce' = map red . (grp . combining) . concat . map (mapper') where
              mapper' :: (Int,String) -> [(String,Int)]
              mapper' = map red . grp . combining .  mapper

-- sample output
-- [("1840",2),("Paris",2),("Zola",2),("born",2),("in",4),("was",2)]
egOutMR :: [(String, Int)]
egOutMR = mapreduce egInput

-- enhanced mapper with local aggregation
-- [[("1840",1),("Paris",1),("Zola",1),("born",1),("in",2),("was",1)],[("1840",1),("Paris",1),("Zola",1),("born",1),("in",2),("was",1)]]
egOutMap' :: [[(String,Int)]]
egOutMap' = map (map red . grp . combining .  mapper) egInput



-- generic map reduce pipeline
simpleMapReduce
    :: ((k1,v1)   -> [(k2,v2)])       -- fonction 'mapper'
    -> ([(k2,v2)] -> [(k2,[v2])])     -- fonction 'sort/group/shuffle'
    -> ((k2,[v2]) -> (k3,v3))         -- fonction 'reduce'
    -> [(k1,v1)]                      -- entrée :  liste de clef valeur
    -> [(k3,v3)]                      -- sortie
simpleMapReduce m c r = map r . c . concat . map m


foldrMapReduce
    :: ((k1,v1)   -> [(k2,v2)])       -- fonction 'mapper'
    -> ([(k2,v2)] -> [(k2,[v2])])     -- fonction 'sort/group/shuffle'
    -> (k2 -> k3)                     -- fonction 'reduce', 1st part : on keys
    -> (v2 -> v3 -> v3)               -- fonction 'reduce', 2st part : on values
    -> v3                             -- fonction 'reduce', default value
    -> [(k1,v1)]                      -- entrée :  liste de clef valeur
    -> [(k3,v3)]                      -- sortie
foldrMapReduce m c r1 r2 z = map (\(x,y) -> (r1 x, foldr r2 z y)) . c . concat . map m


mapreduce'' :: [(Int, String)] -> [(String, Int)]
mapreduce''  = foldrMapReduce mapper (grp . combining) id (+) 0

--Main function takes in A,B and C txt files reads and passes it to map reduce
main :: IO ()
main = do       
       a   <- readFile "A.txt"
       b   <- readFile "B.txt"       
       c   <- readFile "C.txt"
       putStrLn $ show $ simpleMapReduce mapper (grp . combining) red $ [(0,a),(1,b),(2,c)]

